/*
 *
 * Tomasz Jastrzebski gr. I1SID
 *
 * ZPR - Praca Domowa
 * Temat: Funkcja binary_search ze standardowej biblioteki algorithm
 *
 */

/*
 * Funkcja binary_search() jest standardowym szablonem funkcj biblioteki algorithm. Funkcja sprawdza istnienie zadanej wartosci
 * w  okreslonym przedziale kolekcji danych. Do prawidlowego dzialania funkcji niezbedne jest uporzadkowanie elementow w kolekcji
 * (jesli tak nie jest funkcja sie wykona ale nie koniecznie zwroci prawdziwa informacje)
 *
 * Funkcja binary_search() jest implementacja wyszukiwania binarnego opartego o zlozonosci O(logN) co znaczy, ze jest znacznie szybsza od zlozonosci liniowej,
 * dlatego zawsze warto ja stosowac do przeszukiwania uporzadkowanej kolekcji danych jako alternatywa do przeszukiwania linniowego(przeszukiwania calej kolekcji).
 * Dzialanie algorytmu polega na podziale przeszukiwanej kolekcji na pol w kazdej iteracji, dopoki nie bedziemy miec kolekcji jednostkowej lub znajdziemy
 * wyszukiwany element.
 *
 * Istnieja dostepne dwie wersje szablonu funkcji binary_search():
 *
 * template <class ForwardIterator, class LessThanComparable>
 * bool binary_search(ForwardIterator first, ForwardIterator last,
 *                    const LessThanComparable& value);
 *
 * template <class ForwardIterator, class T, class StrictWeakOrdering>
 * bool binary_search(ForwardIterator first, ForwardIterator last, const T& value,
 *                    StrictWeakOrdering comp);
 *
 * Pierwszy szablon wykorzystywany jest dla typow obiektow, ktore maja zadeklarowany operator< i zgodnie z nim uporzadkowane sa elementy
 * w kolekcji. Jesli elementy uporzadkowane sa inaczej lub operator< nie jest znany dla typu nalezy wykorzystac druga wersje funkcji.
 *
 * Parametry wejsciowe funkcji:
 * 		ForwardIterator first - iterator elementu kolekcji od ktorego zaczynamy wyszukiwanie
 * 		ForwardIterator last - iterator elementu kolekcji na ktorym konczymy wyszukiwanie
 * 		const LessThanComparable& value (const T&) - wartosc wyszukiwanego elementu
 * 		StrictWeakOrdering comp - obiekt funkcyjny poruwnujacy dwa elementy ze soba
 *
 * Wynik funkcji:
 * 		true - element zostal znaleziony i istnieje w kolekcji
 * 		false - brak elementu w kolekcji
 *
 *
 */


#include <algorithm>
#include <vector>
#include <iostream>
#include <string>
#include <complex>

using namespace std;	//Algorytm binary_serch() znajduje sie w przestrzeni nazw std

/* PRZYKLADY UZYCIA : */

//Przyklad 1a - wyszukiwanie wartosci w tablicy uporzadkowanych elementow int
void przyklad1a(){

	cout << "PRZYKLAD 1a - WYSZUKIWANIE WARTOSCI W TABLICY ZMIENNYCH TYPU INT UPORZADKOWANYCH ROSNACO" << endl;

	//Tworzenie tablicy elementow uporzadkowanych w kolejnosci rosnacej (zgodnie z operatorem<)
	int tab[]={2,5,7,12,15,16,17,26,47,89};

	int size=sizeof(tab)/sizeof(int);	//Wielkosc tablicy

	//Wywolanie binary_search() dla tablicy - jako iteratory stosujemy wskazniki na elementy tablicy,
	//tzn wskaznikiem poczatka jest nazwa tablicy (a), natomiast iteratorem ostatniego elementu jest a+N
	cout << "Czy wartosc 26 wystepuje w tablicy? (TAK) : " << (binary_search(tab,tab+size,26) ? "TAK" : "NIE") << endl;
	cout << "Czy wartosc 28 wystepuje w tablicy? (NIE): " << (binary_search(tab,tab+size,28) ? "TAK" : "NIE") << endl;
}

//Przyklad 1b - analogicznie jak przyklad 1a, tylko elementy uporzadkowane w kolejnosci rosnacej

//funkcja porownujaca elementy - dziala odwrotnie do operatora <
bool cmp_tab(int a, int b){
	return !(a<=b);
}

void przyklad1b(){

	cout << endl << "PRZYKLAD 1b - WYSZUKIWANIE WARTOSCI W TABLICY ZMIENNYCH TYPU INT UPORZADKOWANYCH MALEJACO" << endl;

	//Tworzenie tablicy elementow uporzadkowanych w kolejnosci malejacej (zgodnie z operatorem>)
	int tab[]={89,47,26,17,16,15,12,7,5,2};

	int size=sizeof(tab)/sizeof(int);	//Wielkosc tablicy

	//Wywolanie binary_search() dla tablicy - jako iteratory stosujemy wskazniki na elementy tablicy,
	//tzn wskaznikiem poczatka jest nazwa tablicy (a), natomiast iteratorem ostatniego elementu jest a+N
	cout << "Czy wartosc 26 wystepuje w tablicy? (TAK) : " << (binary_search(tab,tab+size,26,cmp_tab) ? "TAK" : "NIE") << endl;
	cout << "Czy wartosc 28 wystepuje w tablicy? (NIE): " << (binary_search(tab,tab+size,28,cmp_tab) ? "TAK" : "NIE") << endl;
}


//Przyklad 2 - przeszukiwanie kolekcji z iteratorami
void przyklad2(){

	cout << endl << "PRZYKLAD 2 - PRZESZUKIWANIE KONTENERA VECTOR Z OBIEKTAMI TYPU STRING" << endl;

	//Tworzenie uporzadkowanej kolekcji stringow - obiekty string sa uporzadkowane alfabetycznie i moga zostac wykorzystane
	//poniewaz klasa std::string ma zaimplementowany operator <, ktory poronuje wlasnie ciagi znakow w porzadku alfabetycznym
	vector<string> v;
	v.push_back("abb");
	v.push_back("ag");
	v.push_back("bsd");
	v.push_back("ddd");
	v.push_back("jkl");
	v.push_back("mmm");
	v.push_back("mnn");
	v.push_back("mn1");
	v.push_back("mn2");

	//Wywyolanie funkcji dla kontenera vector - do nawigacji po kolekcji wykorzystujemy standardowe iteratory vectora
	cout << "Czy tekst \"ddd\" wystepuje w tablicy? (TAK) : " << (binary_search(v.begin(),v.end(),"ddd") ? "TAK" : "NIE") << endl;
	cout << "Czy tekst \"mnn\" wystepuje w tablicy? (NIE): " << (binary_search(v.begin(),v.end(),"mnn") ? "TAK" : "NIE") << endl;
}

//Przyklad 3 - przeszukiwanie kontenera vector zawierajacego liczby zespolone posortowane wedlug rosnacych modulow
typedef complex<double> Complex; //Definicja typu liczb zespolonych

//Funkja porownujaca - zwraca wartosc logiczna mowiaca o relacji mniejszosci modulow liczb zespolonych
bool cmp(const Complex& c1, const Complex& c2){
	return abs(c1) < abs(c2);
}

void przyklad3(){

	cout << endl << "PRZYKLAD 2 - PRZESZUKIWANIE KONTENERA VECTOR Z LICZBAMI ZESPOLONYMI O ROSNACYCH MODULACH" << endl;

	//Tworzenie wektora liczb zespolonych o rosnacych modulach
	vector<Complex>v;
	v.push_back(Complex(1,1));
	v.push_back(Complex(2,1));
	v.push_back(Complex(3,2));
	v.push_back(Complex(5,5));
	v.push_back(Complex(15,0));
	v.push_back(Complex(18,1));
	v.push_back(Complex(20,10));

	//Wywolanie funkcji dla powyzszego kontenera
	cout << "Czy wartosc 3+2i wystepuje w tablicy? (TAK) : " << (binary_search(v.begin(),v.end(),Complex(3,2),cmp) ? "TAK" : "NIE") << endl;
	cout << "Czy wartosc 7+2i wystepuje w tablicy? (NIE): " << (binary_search(v.begin(),v.end(),Complex(7,2),cmp) ? "TAK" : "NIE") << endl;
}

//WYWYOLANIE PRZYKLADOW
int main(int argc, char **argv) {
	przyklad1a();
	przyklad1b();
	przyklad2();
	przyklad3();
}

//WYNIK WYKONANIA POWYZSZYCH PRZYKLADOW

//PRZYKLAD 1a - WYSZUKIWANIE WARTOSCI W TABLICY ZMIENNYCH TYPU INT UPORZADKOWANYCH ROSNACO
//Czy wartosc 26 wystepuje w tablicy? (TAK) : TAK
//Czy wartosc 28 wystepuje w tablicy? (NIE): NIE
//
//PRZYKLAD 1b - WYSZUKIWANIE WARTOSCI W TABLICY ZMIENNYCH TYPU INT UPORZADKOWANYCH MALEJACO
//Czy wartosc 26 wystepuje w tablicy? (TAK) : TAK
//Czy wartosc 28 wystepuje w tablicy? (NIE): NIE
//
//PRZYKLAD 2 - PRZESZUKIWANIE KONTENERA VECTOR Z OBIEKTAMI TYPU STRING
//Czy tekst "ddd" wystepuje w tablicy? (TAK) : TAK
//Czy tekst "mnn" wystepuje w tablicy? (NIE): NIE
//
//PRZYKLAD 2 - PRZESZUKIWANIE KONTENERA VECTOR Z LICZBAMI ZESPOLONYMI O ROSNACYCH MODULACH
//Czy wartosc 3+2i wystepuje w tablicy? (TAK) : TAK
//Czy wartosc 7+2i wystepuje w tablicy? (NIE): NIE

